"""Luvdis __version__: A Pure-Python Game Boy Advance Disassembler

Copyright (C) 2020 A. Antonitis. Licensed under the MIT license.
"""
__version__ = '0.5.0'
__doc__ = __doc__.replace('__version__', __version__)
url = __url__ = 'https://github.com/arantonitis/luvdis'
